/* 
** DESCRIPTION: Multiplikation Vektor mit Skalar
** COMPILE:     mpic++ -o vecmult3 vecmult3.cpp
** RUN:         mpiexec -n 2 -machinefile machines $PWD/vecmult3
**
** Diese Version arbeitet mit beliebigen Vektor-Lngen und Scatterv/Gatherv.
*/

#include <iostream>
#include <stdlib.h>
#include <assert.h>
#include <mpi.h>

using namespace std;

int main (int argc, char **argv)
{
	int i;
	int myrank, nprocs;
	int namelen;
	char name[MPI_MAX_PROCESSOR_NAME];

	double *a = NULL, factor;
	int *sizes = NULL;
	int *displs = NULL;

	int n;
	double *local_a;
	int local_n;


	/* MPI initialisieren und Argumente setzen */ 
	MPI_Init(&argc, &argv);

	/* Anzahl der Prozesse erfragen */
	MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

	/* Eigenen Rang erfragen */
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	/* Rechnernamen erfragen */
	MPI_Get_processor_name(name, &namelen);

	/* flush, damit Ausgabe sofort kommt */
	cout << "Process " << myrank << "/" << nprocs << "started on " << name << "\n" << flush;

	
	/* Zahl der Argumente prfen */
	if (argc < 2) {
		if (myrank == 0) {
			cerr << "Usage: mpiexec -n <np> ... vecmult3 <array-len>\n"
				 << "       <array-len>: Length of the array\n";
		}
		MPI_Finalize();
		exit(1);
	}

	/* Achtung: hier fehlt die Prfung des Arguments! */
	n = atoi(argv[1]);


	if (myrank == 0) {
		/* Vektor allokieren. */
		a = new double[n];

		/* Initialisiere a und factor. (Typisch: lesen aus Datei) */
		for (i=0; i<n; i++) {
			a[i] = i/10.0;
		}
		factor = 10.0;

		/* Parameter fr Scatterv / Gatherv allokieren und initialisieren */
		sizes = new int[nprocs];
		displs = new int[nprocs];

		for (i=0; i<nprocs; i++) {
			/*
			** Gleichmige Aufteilung von n Elementen auf nprocs Prozesse.
			** sizes[i] ist die Zahl der Elemente von Proze i,
			** displs[i] ist der Index des ersten Elements fr Proze i.
			** Mit derselben Formel kann auch jeder Proze selbst bestimmen,
			** welchen Teil der globalen Daten er besitzt.
			*/
			sizes[i] = (n + i) / nprocs;
			displs[i] = n / nprocs * i;
			if (i > nprocs - n % nprocs)
				displs[i] += i - (nprocs - n % nprocs);

			cout << "Process " << i << ": displ = " << displs[i] << ", size = " << sizes[i] << "\n";

			/* Sicherstellen, da Aufteilung korrekt ist */
			assert((i==0) || (displs[i] == displs[i-1] + sizes[i-1]));
		}
		assert(displs[i-1] + sizes[i-1] == n);
	}

	
	/* Bestimme Gre des lokalen Teils der Arrays und allokiere diesen */
	local_n = (n + myrank) / nprocs;
	local_a = new double[local_n];

	cout << "Process " << myrank << ": local_n = " << local_n << "\n" << flush;
	fflush(stdout);


	/* Proze 0 sendet factor an alle Prozesse */
	MPI_Bcast(&factor, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

	/* Proze 0 verteilt a auf alle Prozesse */
	MPI_Scatterv(a, sizes, displs, MPI_DOUBLE, local_a, local_n,
				 MPI_DOUBLE, 0, MPI_COMM_WORLD);

	/* Lokalen Teil von a mit factor multiplizieren */
	for (i=0; i<local_n; i++)
		local_a[i] *= factor;

	/* Ergebnis wieder in Proze 0 aufsammeln */
	MPI_Gatherv(local_a, local_n, MPI_DOUBLE, a, sizes, displs,
				MPI_DOUBLE, 0, MPI_COMM_WORLD);

	/*
	** Hinweis: dieses Programm ist natrlich nicht effizient, da zuwenig
	** berechnet wird. Stellen Sie sich aber vor, da statt der Multiplikation
	** eine komplexe Funktion berechnet werden mu, die von a[i] und factor
	** abhngt.
	*/

	if (myrank == 0) {
		/* Ausgabe von a. (Typisch: schreiben in Datei) */
		for (i=0; i<n; i+=n/20) {
			cout << "a[" << i << "] = " << a[i] << "\n";
		}
	}

 
	/* MPI geordnet beenden */
	MPI_Finalize();

	return 0;
}

